//##############################################################################
// Time-stamp: < pdlsthighlight.cpp - 2002/07/01 20:38:21 roland >
//##############################################################################
/** @file

@brief  pdlsthighlight - A tool that can be used as a pdhighlight replacement in
        the ProgDoc tool set (www.progdoc.org).
        Instead of performing syntax highlighting on its own, it wraps the
        extracted code in the TeX environment lstlisting.


@verbatim
+--------------------------------------------------------------------------+
| Roland Weiss                 | Email: weissr@informatik.uni-tuebingen.de |
| University of Tuebingen      |        weissr@uni-tuebingen.de            |
| WWW: http://www-ca.informatik.uni-tuebingen.de/people/weissr/weissr.html |
+--------------------------------------------------------------------------+
@endverbatim */
//##############################################################################


// C/C++ std library
#include <iostream>
#include <fstream>
#include <string>
#include <cstdlib>
#include <string.h>

// command line options parser
#include "options.hpp"
#include "substitute.h"

using namespace std;

// Configuration options.
bool use_colors = true;
bool bold_capname = false;
string language="C++";
bool progDoc = true;

// Variables for command line parameters.
string listing_name;
size_t line_begin = 0, line_end = 0;
string label_name;
string default_label_name, reference_name;
string caption;
string line_sep = "\\pdLineSep", font_size = "\\pdFontSize";
string base_font = "\\pdBaseFont";
string font_enc = "\\pdFontEnc";
string listing_style;
size_t center = 0;


void generateOutput(ifstream &file) {

  // Write prolog.
  cout << "\\begin{flushleft}" << "\\stepcounter{Listing}\n";
  cout << "{\\small";
  if (use_colors)
  {
    cout << "\\color{partfontcol}";
  }

  cout <<  "\n{\\normalfont";
  if (use_colors)
  {
    cout << "\\color{partnamecol}";
  }
  cout << "$\\langle$";
  if (bold_capname)
  {
    cout << "\\textbf{";
  }
  cout << "Listing \\thechapter.\\theListing";
  if (caption.size() > 0)
  {
    cout << ": " << caption;
  }
  if (bold_capname)
  {
    cout << "}";
  }
  cout << "$\\rangle$\\ $\\equiv$}";
  if (label_name.size() > 0)
  {
    cout << "\\mylabel{" << label_name << "}\n";
  }
  if (default_label_name.size() > 0)
  {
    cout << "\\mylabel{" << default_label_name << "}\n";
  }

  cout << "\\begin{lstlisting}[escapeinside={&@!&}@,language=" << language << "]{}\n";

  // Copy extracted code
  string line;
  while (getline(file, line, '\n'))
  {
    cout << line << "\n";
  }

  // Write epilog.
  cout << "\\end{lstlisting}}\n"
       << "\\vspace{0ex}\n"
       << "{\\footnotesize\\addtolength{\\baselineskip}{-1.5ex}\n"
       <<  "Code extracted from file "
       <<  "{";
  if (use_colors)
  {
    cout << "\\color{partnamecol}";
  }
  cout << "\\verb^" << listing_name << "^},";

  if (line_begin != line_end)
  {
    cout <<  " lines " << line_begin << " to " << line_end;
  }
  else
  {
    cout <<  " line " << line_begin;
  }
  if (reference_name.size() > 0)
  {
    cout << ", referenced in listing \\ref{" << reference_name << "}";
  }
  cout << ".}\n";
  cout << "\\end{flushleft}\n";
  cout << "\\noindent\n";
}

void generateProgDocOutput(ifstream &file) {

  cout << "\\ifx\\vhsListing\\undefined\\errmessage{You must include ";
  cout << "progdoc.sty}\\fi\n";
  cout << "\\expandafter\\ifx\\csname indentation\\endcsname\\relax%%\n";
  cout << "\\stepcounter{Listing}\n";
  cout << "\\newlength{\\indentation}\\fi\n{\n";
  cout << "\\ifdim\\pdULdepth>0pt\\ULdepth=\\pdULdepth\\fi%%\n";

  if (listing_name.size() > 0) {
    /* Build up the first Listing title to determine its width */ 
    cout << "\\sbox{\\ListingFirstHead}{ \\small ";
    if (listing_style.size() == 0) cout << "\\uline{";
    cout << "{\\bfseries \\ListingName~";
    if (caption.size() > 0) cout << "\\theListing:} " << caption;
    else cout << "\\theListing:} " << listing_name;
    if (line_begin && line_end && caption.size() == 0)
      cout << "~[\\LineName~" << line_begin << "~\\toName~" << line_end << "]";
    if (reference_name.size() > 0)
      cout << "~(\\ReferenceName~Listing \\ref{" << reference_name 
	   << "} \\PageName~\\pageref{" << reference_name << "})"; 
    if (listing_style.size() == 0) cout << "}";
    cout << "}\n";

    /* Build up the second Listing title to determine its width */ 
    /*
    cout << "\\sbox{\\ListingHead}{ \\small ";
    if (listing_style.size() == 0) cout << "\\uline{";
    cout << "{\\bfseries \\ListingName~";
    if (caption.size() > 0) cout << "\\theListing:} " << caption;
    else cout << "\\theListing:} " << listing_name;
    if (line_begin && line_end && caption.size() == 0)
      cout << "~[\\LineName~" << line_begin << "~\\toName~" << line_end << "]";
    cout << "~(\\ListingContinue)";
    if (listing_style.size() == 0) cout << "}";
    cout << "}\n";
    */
    cout << "\\sbox{\\ListingHead}{}\n";

    /* Determine the width of the widest of the two listing titles, but not */
    /* more then the actual column width */
    cout << "\\settowidth{\\pdCaptionLengthA}{\\usebox{";
    cout << "\\ListingFirstHead}}\n";
    cout << "\\settowidth{\\pdCaptionLengthB}{\\usebox{";
    cout << "\\ListingHead}}\n";
    cout << "\\ifdim\\pdCaptionLengthA<\\pdCaptionLengthB\n";
    cout << "\\let\\pdCaptionLengthA\\pdCaptionLengthB\n";
    cout << "\\fi\n";
    cout << "\\ifdim\\pdCaptionLengthA<\\columnwidth\n";
    cout << "\\let\\pdCaptionLength\\pdCaptionLengthA\n";
    cout << "\\else\\pdCaptionLength\\columnwidth\n";
    cout << "\\fi\n";

    /* Build up the Listing title for the first occurence */ 
    cout << "\\sbox{\\ListingFirstHead}{\\parbox{\\pdCaptionLength}";
    cout << "{\\raggedright\\small ";
    if (listing_style.size() == 0) cout << "\\uline{";
    cout << "{\\bfseries \\ListingName~";
    if (caption.size() > 0) cout << "\\theListing:} " << caption;
    else cout << "\\theListing:} " << listing_name;
    if (line_begin && line_end && caption.size() == 0)
      cout << " [\\LineName~" << line_begin << " \\toName~" << line_end << "]";
    if (reference_name.size() > 0) {
      cout << " (\\ReferenceName{} \\ListingName~\\ref{" << reference_name 
	   << "} \\PageName~\\pageref{" << reference_name << "})";
    }
    if (listing_style.size() == 0) cout << "}";
    cout << "}}\n";

    /* Build up the Listing title for the occurence on subsequent pages */ 
    /*
    cout << "\\global\\sbox{\\ListingHead}{\\parbox{\\pdCaptionLength}";
    cout << "{\\raggedright\\small ";
    if (listing_style.size() == 0) cout << "\\uline{";
    cout << "{\\bfseries \\ListingName~";
    if (caption.size() > 0) cout << "\\theListing:} " << caption;
    else cout << "\\theListing:} " << listing_name;
    if (line_begin && line_end && caption.size() == 0)
      cout << "~[\\LineName~" << line_begin << "~\\toName~" << line_end << "]";
    cout << " (\\ListingContinue)";
    if (listing_style.size() == 0) cout << "}";
    cout << "}}\n";
    */
  }
  cout << "\\usefont{" << font_enc << "}{" << base_font << "}{m}{n}\n";
  cout << "\\fontsize{" << font_size << "}{" << line_sep << "}\\selectfont\n";
  if (default_label_name.size() > 0) 
    cout << "\\mylabel{" << default_label_name << "}\n";
  if (label_name.size() > 0) 
    cout << "\\mylabel{" << label_name << "}\n";

  cout << "\\settowidth{\\indentation}{X}\n";

  if (center) {
    cout << "\\settowidth{\\pdCaptionLengthA}{";
    cout << "\\usebox{\\ListingFirstHead}}\n";
    cout << "\\setlength{\\pdCaptionLengthB}{" << center << "\\indentation}\n";
    cout << "\\ifdim\\pdCaptionLengthA<\\pdCaptionLengthB\n";
    cout << "\\let\\pdCaptionLength\\pdCaptionLengthB\n";
    cout << "\\else\\pdCaptionLength\\pdCaptionLengthA\n";
    cout << "\\fi\n";
    cout << "\\setlength{\\pdCaptionLengthA}{\\pdCaptionLength}";
    cout << "\\addtolength{\\pdCaptionLengthA}{-\\columnwidth}%\n";
    /*
    cout << "\\global\\sbox{\\ListingHead}{\\makebox[\\pdCaptionLength][l]";
    cout << "{\\usebox{\\ListingHead}}}\n";
    cout << "\\smallskip\\noindent%%\n";
    cout << "\\begin{center}\n";
    */
  }
  else {
    cout << "%%\\setlength{\\pdCaptionLength}{" << 80 << "\\indentation}\n";
    /*
    cout << "\\smallskip\\noindent%%\n";
    cout << "\\begin{flushleft}\n";
    */
  }

  if (center) {
    // We need this in order to test if we have version of listings.sty which
    // is higher or lower then 1.0, because the option 'indent' was changed to
    // 'xleftmargin' in ther version 1.0.
    cout << "\\makeatletter%\n";
    cout << "\\setlength{\\@tempdima}{\\lst@version pt}%\n";
    cout << "\\ifdim\\@tempdima<1.0pt%\n";
    cout << "\\providecommand{\\lstIndent}{indent}%\n";
    cout << "\\else%\n";
    cout << "\\providecommand{\\lstIndent}{xleftmargin}%\n";
    cout << "\\fi%\n";
    cout << "\\makeatother%\n";
  }

  cout << "\\begin{lstlisting}[escapeinside={&@!&}@,language=" 
       << language << ","; 
  if (center) cout << "\\lstIndent=-0.5\\pdCaptionLengthA,";
  cout << "keepspaces=true,basewidth=\\indentation,"
       << "commentstyle=\\pdCommentFont,"
       << "keywordstyle=\\pdKeywordFont,stringstyle=\\pdStringFont,"
       << "]{}\n";

  if (listing_name.size() > 0) {
    cout << "&@!&";
    if (center) {
      cout << "\\makebox[\\pdCaptionLength][l]{";
    }
    cout << "\\usebox{\\ListingFirstHead}";
    if (center) {
      cout << "}";
    }
    cout << "\\mbox{}\\\\*[\\medskipamount]\\mbox{}@";
    
    /* Save the Listing title in a mark */ 
    /*
    cout << "\\extramarksPD{\\usebox{\\ListingHead}\\bigskip";
    cout << "}{}%%\n";
    
    cout << "\\makeatletter\\expandafter\\ifx\\csname doublecol@number";
    cout << "\\endcsname\\relax%%\n";
    cout << "\\global\\printcontinuedPDtrue%%\n";
    cout << "\\afterpage{\\ifprintcontinuedPD\\noindent";
    cout << "\\topxmarkPD\\fi\\global\\printcontinuedPDfalse}%%\n";
    cout << "\\else\\ifnum\\doublecol@number=0%%\n";
    cout << "\\global\\printcontinuedPDtrue%%\n";
    cout << "\\afterpage{\\ifprintcontinuedPD\\noindent";
    cout << "\\topxmarkPD\\fi\\global\\printcontinuedPDfalse}%%\n";
    cout << "\\fi\\fi\\makeatother%%\n";
    */
  }

  // Copy extracted code
  string line;
  while (getline(file, line, '\n'))
  {
    cout << line << "\n";
  }

  // Write epilog.
  cout << "\\end{lstlisting}\n";
  
  /*
  cout << "\\extramarksPD{}{}\n";
  if (center) {
    cout << "\\end{center}\n";
  }
  else {
    cout << "\\end{flushleft}\n";
  }
  */
  cout << "\n% We need a blank line if we want that baselineskip will be set ";
  cout << "correctly\n}\n";
 
}

//##############################################################################

char *cvs_version = "$Revision: 1.9 $";
char version_string[250] = { '\0' };
char *cvs_release = "$Name:  $";
char release_string[250] = "Unreleased";

int main(int argc, char** argv)
{

  makeVersionNr(version_string, cvs_version);
  makeReleaseNr(release_string, cvs_release);

  // Make sure that at least the name of the input file is present.
  if (argc < 2)
  {
    cout << "pdlsthighlight" << version_string << " (ProgDoc Rel. "
	 << release_string << ")\n\n";
    cout << "usage: pdlsthighlight [options] <input_fname>\n";
    return EXIT_FAILURE;
  }

  // Specify the command line options.
  const char* optv[] = {
    "N:Name <string>", // name of the source file
    "b:begin <number>", // begin line nr
    "e:end <number>", // end line nr
    "L:Label <string>", // user provided label
    "D:GenLabel <string>", // label generated by ProgDoc
    "R:GenRef <string>", // reference generated by ProgDoc
    "A:Caption <string>", // user provided caption
    "n| ", // nuWeb Listing format. (ProgDoc listing format is the default.)
    "m:language <string>", // language of the source code
    "l:line-sep <string>", // line separation
    "s:font-size <string>", // base font size
    "E:font-encoding <string>", // TeX encoding of the base font (default OT1)
    "F:base-font <string>", // base font name
    "W:caption-style <string>", // underline/noUnderline. Defualt us underline.
    "z?center <number>", // Takes the max line width as optional argument.
    "v| ", // print version
    "t| ", // always generated by pdweave, to supress warning
    "w| ", // Ignore. (Generated by pdweave to use longtable)
    "u| ", // Ignore. (Generated by pdweave to underline keywords)
    0
  };

  // Prepare oprions parser.
  Options opts(*argv, optv);
  OptArgvIter it(--argc, ++argv);

  // Iterate through command line parameters.
  const char * optarg;
  while (char optchar = opts(it, optarg))
  {
    switch (optchar)
    {
    case 'N':
      listing_name = substitute(optarg);
      break;
    case 'A':
      caption = optarg;
      break;
    case 'L':
      label_name = optarg;
      break;
    case 'D':
      default_label_name = optarg;
      break;
    case 'R':
      reference_name = optarg;
      break;
    case 'e':
      if (optarg != 0) line_end = atoi(optarg);
      break;
    case 'b':
      if (optarg != 0) line_begin = atoi(optarg);
      break;
    case 'm':
      if (optarg != 0) language = optarg;
      break;
    case 'n':
      progDoc = false;
      break;
    case 'l':
      if (optarg != 0) line_sep = optarg;
      break;
    case 's':
      if (optarg != 0) font_size = optarg;
      break;
    case 'E':
      if (optarg != 0) font_enc = optarg;
      break;
    case 'F':
      if (optarg != 0) base_font = optarg;
      break;
    case 'W':
      if (optarg != 0) listing_style = optarg;
      break;
    case 'z':
      if (optarg != 0) center = atoi(optarg); else center = 80;
      break;
    case 'v':
      cout << "pdlsthighlight" << version_string << " (ProgDoc Rel. "
           << release_string << ")\n\n";
      return EXIT_SUCCESS;
      break;
    }
  }

  // Open the input file, which is given as last command line parameter.
  string fname = argv[argc - 1];
  ifstream tmp_file(fname.c_str());

  if (progDoc) {
    generateProgDocOutput(tmp_file);
  }
  else {
    generateOutput(tmp_file);
  }
}

// Now obsolete, because pdweave creates the correct escape codes.
//     // Convert pdhighlight escapes to lstlisting escapes.
//     if (line.find("/*?") != string::npos)
//     {
//       line.replace(line.find("/*?"), 3, "&@!&");
//       cout << line << "\\hbox{";
//       if (use_colors)
//       {
//         cout << "\\color{partnamecol}";
//       }
//       cout << "\n";
//     }
//     else if (line.find("?*/") != string::npos)
//     {
//       cout << "}@" << "\n";
//     }
